
#include <linux/module.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/kernel.h>
#include <linux/proc_fs.h>
#include <linux/init.h>
#include <asm/irq.h>

#include <asm/segment.h>
#include <asm/hardware.h>

#include "gpio_example.h"

// #define DEBUG

/* Administratrivia */

#define	IRQ_GPIO_11_27(x)	(32 + (x) - 11)
#define SA1100_GPIO_TO_IRQ(i)	(((i) < 11) ? (i) : IRQ_GPIO_11_27(i))
#define SA1100_IRQ_TO_GPIO(i) 	(((i) < 11) ? (i) : GPIO_11_27_IRQ(i))

#ifndef GPIO_EXAMPLE_MAJOR
#define GPIO_EXAMPLE_MAJOR 61
#endif

static struct fasync_struct *fasync[GPIO_MAX+1];

static void set_gpiobit(unsigned int minor)
{
#ifdef DEBUG
  printk(KERN_DEBUG "in set_gpiobit\n");
#endif
	GPSR = GPIO_GPIO(minor);
}

static void clear_gpiobit(unsigned int minor)
{
#ifdef DEBUG
  printk(KERN_DEBUG "in clear_gpiobit\n");
#endif
	GPCR = GPIO_GPIO(minor);
}

static int check_gpio_gafr(unsigned int minor)
{
        return (GAFR & GPIO_GPIO(minor)) ? 1 : 0;
}

static int check_gpio_gpdr(unsigned int minor)
{
        return (GPDR & GPIO_GPIO(minor)) ? 1 : 0;
}

static ssize_t gpio_example_write(struct file * file, const char * buf, size_t count, loff_t * ppos)
{
	struct inode *inode = file->f_dentry->d_inode;
	unsigned int minor = MINOR(inode->i_rdev);
 
#ifdef DEBUG
	printk(KERN_DEBUG "in gpio_example_write: minor=%d, buf=%#x\n", minor, *buf);
#endif

	if (!check_gpio_gpdr(minor)) {
#ifdef DEBUG
	  printk(KERN_DEBUG "gpio_write but GPIO(%d) is not in output mode.\n", minor);
#endif
	  return -EIO;
	}

	if (check_gpio_gafr(minor)) {
#ifdef DEBUG
	  printk(KERN_DEBUG "gpio_write but GPIO(%d) has alternate function set.\n", minor);
#endif
	  return -EIO;
	}

	if      ( (*buf) == '1') set_gpiobit(minor);
	else if ( (*buf) == '0') clear_gpiobit(minor);
	return(1);
}

static ssize_t gpio_example_read(struct file * file, char * buf, size_t count, loff_t * ppos) 
{
        struct inode *inode = file->f_dentry->d_inode;
	unsigned int minor = MINOR(inode->i_rdev);
	int bitstate;

	if (check_gpio_gafr(minor)) {
#ifdef DEBUG
	  printk(KERN_DEBUG "gpio_write but GPIO(%d) has alternate function set.\n", minor);
#endif
	  return -EIO;
	}

        bitstate = (GPLR & GPIO_GPIO(minor)) >> minor;

        *buf = bitstate ? '1' : '0';

	return 1;
}


static void gpio_example_irq_handler(int irq, void *dev_id, struct pt_regs *regs )
{
        struct fasync_struct *this_fasync;
	int gpio;

	gpio = SA1100_IRQ_TO_GPIO(irq);

	if (gpio < 0 || gpio > GPIO_MAX)
	  printk(KERN_WARNING "gpio_example: invalid gpio: %d from irq: %x\n", gpio, irq);

	this_fasync = fasync[gpio];

	if (this_fasync)
	  kill_fasync(&this_fasync, SIGIO, POLL_IN);
}

static int gpio_example_register_irq(unsigned int minor)
{
        if (minor < 0 || minor > GPIO_MAX) {
#ifdef DEBUG
             printk("gpio_example_register_irq: minor: %d out of range\n", minor);
#endif
             return -EINVAL;
	}

      	set_GPIO_IRQ_edge(GPIO_GPIO(minor), GPIO_RISING_EDGE);

	if (request_irq(SA1100_GPIO_TO_IRQ(minor), gpio_example_irq_handler, 0, "GPIO_EXAMPLE", NULL)) { 
             printk(KERN_WARNING "gpio_example.o: Could not allocate IRQ for GPIO%d.\n", minor);
	     return -EINVAL;
	}

	return 0;
}

static int gpio_example_gafr(int value, unsigned int minor)
{
        if (value)
        	  GAFR |= GPIO_GPIO(minor);
	else
        	  GAFR &= ~GPIO_GPIO(minor);
	return 0;
}

static int gpio_example_gpdr(int value, unsigned int minor)
{
        if (value)
        	  GPDR |= GPIO_GPIO(minor);
	else
        	  GPDR &= ~GPIO_GPIO(minor);
	return 0;
}

static int gpio_example_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
	unsigned int minor = MINOR(inode->i_rdev);

#ifdef DEBUG
	printk(KERN_DEBUG "gpio_example: ioctl: %d, minor: %d\n", cmd, minor);
#endif

	switch (cmd) {

	case GPIO_EXAMPLE_SET_GAFR:
		return gpio_example_gafr(1, minor);

	case GPIO_EXAMPLE_CLEAR_GAFR:
		return gpio_example_gafr(0, minor);

	case GPIO_EXAMPLE_GPDR_INPUT:
		return gpio_example_gpdr(0, minor);

	case GPIO_EXAMPLE_GPDR_OUTPUT:
		return gpio_example_gpdr(1, minor);

	default:
                return -ENOTTY;
	}
}

static int gpio_example_fasync(int fd, struct file *file, int on)
{
	struct inode *inode = file->f_dentry->d_inode;
	unsigned int minor = MINOR(inode->i_rdev);
	int retval;

#ifdef DEBUG
	printk(KERN_DEBUG "in gpio_example_fasync\n");
#endif

	if (check_gpio_gpdr(minor)) {
#ifdef DEBUG
	  printk(KERN_DEBUG "gpio_fasync but GPIO(%d) is not in input mode.\n", minor);
#endif
	  return -EIO;
	}

	if (check_gpio_gafr(minor)) {
#ifdef DEBUG
	  printk(KERN_DEBUG "gpio_fasync but GPIO(%d) has alternate function set.\n", minor);
#endif
	  return -EIO;
	}

	retval = fasync_helper(fd, file, on, &fasync[minor]);
	if (retval < 0) {
                printk(KERN_DEBUG "fasync_helper failed: %d\n", retval);
		return retval;
	}

	return gpio_example_register_irq(minor);
}

static int gpio_example_fasync_release(int fd, struct file *file, int on)
{
	struct inode *inode = file->f_dentry->d_inode;
	unsigned int minor = MINOR(inode->i_rdev);
	int retval;

#ifdef DEBUG
	printk(KERN_DEBUG "in gpio_example_fasync_release\n");
#endif

	fasync[minor] = 0;

	retval = fasync_helper(fd, file, on, &fasync[minor]);
	if (retval < 0) {
                printk(KERN_DEBUG "fasync_helper failed: %d\n", retval);
		return retval;
	}

#ifdef DEBUG
	printk(KERN_DEBUG "gpio_example_release: freeing irq: %d, minor: %d\n", SA1100_GPIO_TO_IRQ(minor), minor);
#endif

	free_irq(SA1100_GPIO_TO_IRQ(minor), NULL);

	return retval;
}

int gpio_example_release(struct inode *inode, struct file *file)
{
	unsigned int minor = MINOR(inode->i_rdev);

	if (fasync[minor])
          return gpio_example_fasync_release(-1, file, 0);
	else
	  return 0;
}

static struct file_operations gpio_example_fops = 
{
	read:	 gpio_example_read,
	write:   gpio_example_write,
	ioctl:   gpio_example_ioctl,
	fasync:  gpio_example_fasync,
	release: gpio_example_release,
};

static struct proc_dir_entry *proc_gpio_example;

static int gpio_example_state_proc(char *buf, char **start, off_t offset, int len)
{
	char	*p = buf;
	int i;

	p += sprintf(p, "Alternative Function GPIOs:");
	for (i = 0; i < GPIO_MAX; i++) {
            if (GAFR & GPIO_GPIO(i)) {
	         p += sprintf(p, " %2d", i);
	    }
	}

	p += sprintf(p, "\nInput GPIOs: ");
	for (i = 0; i < GPIO_MAX; i++) {
            if (!(GAFR & GPIO_GPIO(i)) && !(GPDR & GPIO_GPIO(i))) {
	      p += sprintf(p, " %2d", i);
	    }
	}
	p += sprintf(p, "\nInput values:");
	for (i = 0; i < GPIO_MAX; i++) {
            if (!(GAFR & GPIO_GPIO(i)) && !(GPDR & GPIO_GPIO(i))) {
	      p += sprintf(p, " %2d", (GPLR & GPIO_GPIO(i)) >> i);
	    }
	}

	p += sprintf(p, "\nOutput GPIOs: ");
	for (i = 0; i < GPIO_MAX; i++) {
            if (!(GAFR & GPIO_GPIO(i)) && (GPDR & GPIO_GPIO(i))) {
	      p += sprintf(p, " %2d", i);
	    }
	}
	p += sprintf(p, "\nOutput values:");
	for (i = 0; i < GPIO_MAX; i++) {
            if (!(GAFR & GPIO_GPIO(i)) && (GPDR & GPIO_GPIO(i))) {
	      p += sprintf(p, " %2d", (GPLR & GPIO_GPIO(i)) >> i);
	    }
	}

	p += sprintf(p, "\n");

	return p - buf;
}

int __init init_gpio_example(void)
{
	if (register_chrdev(GPIO_EXAMPLE_MAJOR,"gpio_example",&gpio_example_fops)) {
		printk("gpio_example: unable to get major %d\n", GPIO_EXAMPLE_MAJOR);
		return -EIO;
	}
	else {
#ifdef CONFIG_PROC_FS
		proc_gpio_example = proc_mkdir("driver/gpio_example", NULL);
		proc_gpio_example->owner = THIS_MODULE;
		create_proc_info_entry("status", 0, proc_gpio_example, gpio_example_state_proc);
#endif
		printk("ADS gpio_example driver initialized\n");
	}

	return 0;
}


void __init unregister_gpio_example_driver(void)
{
#ifdef CONFIG_PROC_FS
	remove_proc_entry("status", proc_gpio_example);
	remove_proc_entry("driver/gpio_example", NULL);
#endif
	unregister_chrdev(GPIO_EXAMPLE_MAJOR,"gpio_example");
}

module_init(init_gpio_example);
module_exit(unregister_gpio_example_driver);
